// Auf 8 Bit optimierte Entschlsselung der DCF77-Wetterdaten.
// Quelle: meteocrypt_working.c by Walter Freywald (c) 2011
// Umsetzung in C# durch DCF77Freak (c) 2011
// Umsetzung zurck in C und DWORD-Operationen (nur Dekoder) 120825
// Umsetzung auf 8-Bit-Assemblernhe zum Test 240612
#ifdef _DEBUG
#include <windows.h>
#include <string.h>
#pragma intrinsic(memcpy)
#else
// wirklich: Kein Include erforderlich!
typedef unsigned long DWORD;
#endif

typedef unsigned char byte;

byte cipher_key[10];

#if 0
void _declspec(naked) _cdecl _aullshr() {_asm{
	btr	ecx,5
	jc	short l1
	shrd	eax,edx,cl
	shr	edx,cl
	ret
l1:	mov	eax,edx
	xor	edx,edx
	shr	eax,cl
	ret 
}}
#endif

// Uhrzeit-Teil (hintere 5 Byte) in zwei 20-Bit-Stcken rechts rotieren
static void _declspec(naked) _fastcall Rotate2020R() {
 _asm{	lea	ecx,byte ptr[cipher_key]
 	mov	ah,byte ptr[ecx+9]
  	shr	ah,1
	rcr	dword ptr[ecx+5],1	// in AVR-Assembler auflsen
	mov	al,byte ptr[ecx+7]
	jnc	l2
	test	al,0x08			// in AVR-Assembler hilft bst und bld!
	jz	l1
	or	ah,0x80
l1:	or	al,0x08
	jmp	short l4
l2:	test	al,0x08
	jz	l3
	or	ah,0x80
l3:	and	al,~0x08
l4:	mov	byte ptr[ecx+9],ah
	mov	byte ptr[ecx+7],al
	ret
 }
}

// Datenteil Hi-20-Bit zum Lo-20-Bit schieben (in avr-Assembler mit swap)
static void _declspec(naked) _fastcall Shift20R() {
 _asm{	lea	ecx,byte ptr[cipher_key]
	mov	ah,byte ptr[ecx+3]
	ror	ah,4		// swap
	mov	al,byte ptr[ecx+2]
	ror	al,4		// swap
	and	al,15
	xor	al,ah
	and	ah,15
	xor	al,ah
	mov	byte ptr[ecx+0],al
	mov	al,byte ptr[ecx+4]
	ror	al,4		// swap
	xor	ah,al
	and	al,15
	xor	ah,al
	mov	byte ptr[ecx+1],ah
	mov	byte ptr[ecx+2],al
	mov	byte ptr[ecx+3],0
	mov	byte ptr[ecx+4],0
	ret
 }
}

// Uhrzeit-Schlssel (Low und High getrennt) rechts rotieren
__forceinline void ShiftTimeRight(byte round) {
 byte count=1;
 switch (round) {
  case 16:
  case 8:
  case 7:
  case 3: count++;
 };
 do Rotate2020R(); while (--count);
}

DWORD pick30bits(const byte X[30]) {
 DWORD ret = 0;
 int i;
 for (i=0; i < 30; i++) {
  byte b = X[i];
  if (cipher_key[b>>3] & 1<<(b&7)) ret |= 1<<30;
  ret>>=1;
 }
 return ret;
}

// 20 auf 30 Bits expandieren
__forceinline DWORD ExpandR() {
 static const byte X[30]={
  20,21,22,23,39,24,
  24,25,26,27,23,28, 
  28,29,30,31,27,32,
  32,33,34,35,31,36, 
  36,37,38,39,35,20};
 return pick30bits(X);
}

// 40 Bits (Zeit) auf 30 Bits komprimieren
__forceinline DWORD CompressKey() {
 static const byte X[30]={	// 40..79
  49,45,65,75,55,46,
  77,47,70,64,52,50, 
  66,76,56,74,40,71,
  62,44,61,59,43,41, 
  54,78,57,60,58,68};
 return pick30bits(X);
}

// 30 (5x6) Bits auf 20 (5x4) Bits komprimieren (Scrambling-Box)
// Anschlieend 20 Bits in 20 Bits verwrfeln (Permutations-Box)
__forceinline DWORD Sbox_Pbox(DWORD src) {
 static const unsigned __int64 S[5*4]={
  0x6A4C8190D7F352EB,0xE6F9075C23A1B8D4,0x5E92B684FC0A3D71,0x79854A21ECB3360F,
  0x54813AE96FD7C20B,0x437EFAC159BD0286,0x459D2F8160EAB37C,0x625A3C48EF0BD971,
  0x0A239817FE654CDB,0xEC46F0A3D17982B5,0xE410D9A25B68F3C7,0x6934DCB5FA28071E,
  0x2BC85D1940E67F3A,0x8AB62EF970345D1C,0xA170D9283E65F4CB,0x2E803C6D9A517FB4,
  0xE4195BC38D76F02A,0x34A01B768FECD592,0xE27A594B63C1FD08,0x5472A8E1BF69C0D3};
 static const byte P[20] =
  {18,13,15,6,8,4,10,7,2,1,19,9,14,17,12,3,11,5,0,16};
 DWORD ret=0;
 byte h[3];
// 5 Runden (5 x 6 bit -> 5 x 4 bit komprimieren)
// Box-Zugehrigkeit <src>: --444444 33333322 22221111 11000000
// Box-Zugehrigkeit <ret>: -------- ----4444 33332222 11110000
 byte i,na,n;
 h[2]=0;
 for (i=0;;) {
  na = (byte)src&0x3F;	// Nibble-Adresse
  n = ((byte*)S)[(byte)(i|(na>>1))];
  n = na&1 ? n>>4 : n&15;
  if (i&1<<5) n<<=4;
  h[2] |= n;
  if (i&1<<5) {h[0]=h[1]; h[1]=h[2]; h[2]=0;}
  if ((i+=1<<5)==5<<5) break;
  src>>=6;
 }
 for (i=0; i < 20; i++) {
  byte b = P[i];
  if (h[b>>3] & 1<<(b&7)) ret |= 1<<20;
  ret>>=1;
 }
 return ret;
}

// Ermittelt Wetterdaten aus den entschlsselten Bytes und prft auf Korrektheit.
// <cipher_key> = entschlsselte Bytes (2 x 20 Bits)
// (C = CRC, D = Nutzdaten, d = angeblich nur 2 Nutzdatenbits): DDDDD:CCCCd
// <ret> Gibt die aus dem Klartextbytes ermittelten Wetterdaten zurck,
//	 also dDDDDD; -1 bei Fehler (anders als der HKW581-Mikrocontroller)
__forceinline long GetMeteoFromPlain() {
 byte lonibble = cipher_key[0]&0x0F;
 if (*(DWORD*)cipher_key&0xFFFF0  != 0x25010) return -1;	// negativ: Fehler (besser als Kode 0x20001)
 Shift20R();
 cipher_key[2]|=lonibble<<4;
 return *(DWORD*)cipher_key;	// 22 (24) Nutzbits
}

// Entschlsselt einen Datensatz.
// <cipher_key> = Verschlsselte Wetterdaten (80 Bit) in binrer Form, Klartext-Uhrzeit enthaltend
// Die Uhrzeit muss eine derartige Minuten-Nummer enthalten, deren Teilung durch 3 den Rest 2 lsst.
// Wird gebndelt aus 3 Minuten, beginnend bei einer vollen Stunde, also anfangs mit den Minuten-Nummern 1,2,3;
// der Minutenwert entspricht also der "mittleren" Minute.
// Aus einem 82-Bit-Datenstrom (fr den HKW581) werden 80 Bit durch Weglassen von Bit 0 (erstes Bit) und Bit 7.
// <cipher_key> enthlt bei Erfolg am Ende in den unteren 3 Byte die entschlsselten Wetterdaten
// in binrer Form (untere 22 Bit + 2 Bit Status).
static long DecodeData() {
 byte i = 16;		// Rundenzhler
 do{
  DWORD key;	// 30-bit-Schlssel (aus Uhrzeit gebildet, in jeder Runde wechselnd)
  ShiftTimeRight(i);
  key = ExpandR() ^ CompressKey();	// expR XOR komprimierter Schlssel
  key = *(DWORD*)cipher_key&0xFFFFF ^ Sbox_Pbox(key);	// L XOR P-boxed Rundenschlssel (<key> jetzt L')
  Shift20R();				// L = R: tauschen
  *(DWORD*)(cipher_key+1) |= key<<12;	// R = L'
 }while(--i);
 return GetMeteoFromPlain();	// Klartext herausholen
}

#ifdef _DEBUG
// Prfen der Dekoder-Funktion
int CALLBACK WinMainCRTStartup(HINSTANCE a, HINSTANCE b, char* c, int d) {
 const BYTE testdata[10]={0x80,0x24,0xCA,0x7B,0x27,0x38,0x02,0x20,0x88,0x15};
 memcpy(cipher_key,testdata,10);
 return DecodeData()==0x0C1011;
}
#endif
