// Funktionen zum Entschlüsseln der via DCF77 übertragenen Wetterdaten.
// Quelle: meteocrypt_working.c by Walter Freywald (c) 2011
// Umsetzung in C# durch DCF77Freak (c) 2011
// Umsetzung zurück in C und DWORD-Operationen (nur Dekoder) 120825

#include "FunkUsb.h"
#include <avr/pgmspace.h>

void Rotate2020R() {
 asm(

"	lsr	%2	\n"
"	ror	%1	\n"
"	ror	%0	\n"
:"+r"(b[0]),"+r"(b[1]),"+r"(b[2]):);
}

// 20 Bit rechts rotieren (beliebig weit, count = 1..19)
void B20::rotateR(byte count) {
 asm(
"0:	lsr	%2	\n"
"	ror	%1	\n"
"	ror	%0	\n"
"	brcc	1f	\n"
"	ori	%2,8	\n"
"1:	dec	%3	\n"
"	brne	0b	\n"
:"+r"(b[0]),"+r"(b[1]),"+d"(b[2]):"r"(count));
}

struct B40{
 B20 lo,hi;
 operator byte*() {return lo;}
 operator const byte*() const {return lo;}
};

// Uhrzeit-Schlüssel (Low und High getrennt) rechts rotieren
inline void ShiftTimeRight(B40*src, int round) {
 byte count=1;
 switch (round) {
  case 16:
  case 8:
  case 7:
  case 3: count++;
 }
// "Uhrzeit" (2 x 20 Bit) jeweils rechts rotieren
 src->lo.rotateR(count);
 src->hi.rotateR(count);
}

// 20 auf 30 Bits expandieren
inline DWORD ExpandR(const B20&src) {
 static const PROGMEM byte X[30]={
  0, 1, 2, 3,19, 4,
  4, 5, 6, 7, 3, 8, 
  8, 9,10,11, 7,12,
 12,13,14,15,11,16, 
 16,17,18,19,15, 0};
 DWORD ret = 0;
 for (byte i=0; i < 30; i++) {
  byte b = pgm_read_byte(X+i);
  if (src[b>>3] & 1<<(b&7)) ret |= 1UL<<30;
  ret>>=1;
 }
 return ret;
}

// 40 auf 30 Bits komprimieren
inline DWORD CompressKey(const B40&Time) {
 static const PROGMEM byte X[30]={	// nur 0..19 und 32..51
 0x09, 0x05, 0x25, 0x2F, 0x0F, 0x06,
 0x31, 0x07, 0x2A, 0x24, 0x0C, 0x0A, 
 0x26, 0x30, 0x10, 0x2E, 0x00, 0x2B,
 0x22, 0x04, 0x21, 0x13, 0x03, 0x01, 
 0x0E, 0x32, 0x11, 0x20, 0x12, 0x28};
 DWORD ret=0;
 for (byte i=0; i < 30; i++) {
  byte b = pgm_read_byte(X+i);
  if (b>=32) b-=8;
  if (Time[b>>3] & 1<<(b&7)) ret |= 1UL<<30;
  ret>>=1;
 }
 return ret;
}

// 30 (5x6) Bits auf 20 (5x4) Bits komprimieren (Scrambling-Box)
static B20 Sbox(DWORD src) {
 static const PROGMEM QWORD X[5*4]={
  0x6A4C8190D7F352EB,0xE6F9075C23A1B8D4,0x5E92B684FC0A3D71,0x79854A21ECB3360F,	// Nibble 0
  0x54813AE96FD7C20B,0x437EFAC159BD0286,0x459D2F8160EAB37C,0x625A3C48EF0BD971,
  0x0A239817FE654CDB,0xEC46F0A3D17982B5,0xE410D9A25B68F3C7,0x6934DCB5FA28071E,
  0x2BC85D1940E67F3A,0x8AB62EF970345D1C,0xA170D9283E65F4CB,0x2E803C6D9A517FB4,
  0xE4195BC38D76F02A,0x34A01B768FECD592,0xE27A594B63C1FD08,0x5472A8E1BF69C0D3};
 B20 ret;
 ret.b[2]=0;
// 5 Runden (5 x 6 bit -> 5 x 4 bit komprimieren)
// Box-Zugehörigkeit <src>: --444444 33333322 22221111 11000000
// Box-Zugehörigkeit <ret>: -------- ----4444 33332222 11110000
 for (byte i=0;;) {
  byte na = src&0x3F;	// Nibble-Adresse
  byte n = pgm_read_byte((byte*)X+byte(i|na>>1));
  asm(
"	sbrc	%1,0	\n"
"	 swap	%0	\n"
"	andi	%0,15	\n"
"	sbrc	%2,5	\n"
"	 swap	%0	\n"
:"+d"(n):"r"(byte(src)),"r"(i));
//  n = na&1 ? n>>4 : n&15;
//  if (i&0x20) n<<=4;
  ret.b[2] |= n;
  if (i&0x20) {ret.b[0]=ret.b[1]; ret.b[1]=ret.b[2]; ret.b[2]=0;}
  if ((i+=0x20)==5<<5) break;
  src>>=6;
 }
 return ret;
}

// 20 Bits in 20 Bits verwürfeln (Permutations-Box)
static B20 Pbox(const B20&src) {
 static const PROGMEM byte X[20] =
  {18,13,15,6,8,4,10,7,2,1,19,9,14,17,12,3,11,5,0,16};
 B20 ret {};
 for (byte i=0; i < 20; i++) {
  byte b = pgm_read_byte(X+i);
  if (src[b>>3] & 1<<(b&7)) ret.b[2] |= 16;
  ret.shiftR();
 }
 return ret;
}
// _bittest(&src,...) würde hier "&0x1F" erfordern; das macht den Kode größer.
// Gäbe es intrinsisches Register-"bt", wäre dieser Schritt nicht erforderlich.


// Ermittelt Wetterdaten aus den entschlüsselten Bytes und prüft auf Korrektheit.
// <plain> = entschlüsselte Bytes (2 x 20 Bits)
// (C = CRC, D = Nutzdaten, d = angeblich nur 2 Nutzdatenbits): 000DDDDD 000CCCCd
// <ret> Gibt die aus dem Klartextbytes ermittelten Wetterdaten zurück,
//	 also dDDDDD; -1 bei Fehler (anders als der Mikrocontroller)
long GetMeteoFromPlain(B40 plain) {
 if (plain.lo.b[2]!=2 || plain.lo.b[1]!=0x50 || plain.lo.b[0]&0xF0!=0x10) return -1;	// negativ: Fehler (günstiger als 0x20001)
 return (long)plain.hi[2]<<16 | plain.hi[1]<<8 | plain.hi[0] | (long)(plain.lo[0]&0x0F)<<20;	// 22 (24) Nutzbits
}

// Verteilt 40 Bit (5 Byte) auf 48 Bit (6 Byte); Praxistest!
static void LoadCipher(B40&dst,const byte*src) {
 dst.lo.b[0] = src[0];			// Cipher.L: 20 untere Bits
 dst.lo.b[1] = src[1];
 dst.lo.b[2] = src[2]&0x0F;
 dst.hi.b[0] = src[2]>>4|src[3]<<4;	// Cipher.R: 20 obere Bits
 dst.hi.b[1] = src[3]>>4|src[4]<<4;
 dst.hi.b[2] = src[4]>>4;
}

// Entschlüsselt einen Datensatz.
// <cipher_key> = Verschlüsselte Wetterdaten (80 Bit) in binärer Form, Klartext-Uhrzeit enthaltend
// Die Uhrzeit muss eine derartige Minuten-Nummer enthalten, deren Teilung durch 3 den Rest 2 lässt.
// Wird gebündelt aus 3 Minuten, beginnend bei einer vollen Stunde, also anfangs mit den Minuten-Nummern 1,2,3;
// der Minutenwert entspricht also der "mittleren" Minute.
// Aus einem 82-Bit-Datenstrom (für den HKW581) werden 80 Bit durch Weglassen von Bit 0 (erstes Bit) und Bit 7.
// <ret> Gibt die entschlüsselten Wetterdaten in binärer Form (untere 22 Bit + 2 Bit Status) zurück.
long DecodeData() {
 B40 cipher;	// erst Chiffre, dann Klartext (2 x 20 bit)
 B40 time;	// 40-Bit-Uhrzeitregister (2 x 20 bit)
 byte i = 16;		// Rundenzähler

 LoadCipher(cipher,cipher_key+0);
 LoadCipher(time,  cipher_key+5);
 do{
  ShiftTimeRight(&time, i);
// 30-bit-Schlüssel (aus Uhrzeit gebildet, in jeder Runde wechselnd)
  DWORD K = ExpandR(cipher.hi) ^ CompressKey(time);	// expR XOR komprimierter Schlüssel
  B20 L = Sbox(K);
  L = cipher.lo ^ Pbox(L);	// L XOR P-boxed Rundenschlüssel
  cipher.lo = cipher.hi;	// L = R: tauschen
  cipher.hi = L;		// R = L'
 }while(--i);
 return GetMeteoFromPlain(cipher);	// Klartext am Ende interpretieren, ggf. Fehlerkorrektur
}
